/* GStreamer
 * Copyright (C) 2021 Seungha Yang <seungha@centricular.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#include <gst/gst.h>
#include <windows.h>
#include <mfapi.h>

G_BEGIN_DECLS

gboolean gst_mf_plat_load_library (void);

HRESULT __stdcall GstMFTEnum2 (GUID guidCategory,
                               UINT32 Flags,
                               const MFT_REGISTER_TYPE_INFO * pInputType,
                               const MFT_REGISTER_TYPE_INFO * pOutputType,
                               IMFAttributes * pAttributes,
                               IMFActivate *** pppMFTActivate,
                               UINT32 * pnumMFTActivate);

HRESULT __stdcall GstMFCreateDXGIDeviceManager (UINT * resetToken,
                                                IMFDXGIDeviceManager ** ppDeviceManager);

HRESULT __stdcall GstMFCreateVideoSampleAllocatorEx (REFIID riid,
                                                     void** ppSampleAllocator);

G_END_DECLS
