/* GStreamer
 * Copyright (C) 2013 David Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef _GST_VIDEO_DIFF_H_
#define _GST_VIDEO_DIFF_H_

#include <gst/video/video.h>
#include <gst/video/gstvideofilter.h>
#include <string.h>

G_BEGIN_DECLS

#define GST_TYPE_VIDEO_DIFF   (gst_video_diff_get_type())
#define GST_VIDEO_DIFF(obj)   (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_VIDEO_DIFF,GstVideoDiff))
#define GST_VIDEO_DIFF_CLASS(klass)   (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_VIDEO_DIFF,GstVideoDiffClass))
#define GST_IS_VIDEO_DIFF(obj)   (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_VIDEO_DIFF))
#define GST_IS_VIDEO_DIFF_CLASS(obj)   (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_VIDEO_DIFF))

typedef struct _GstVideoDiff GstVideoDiff;
typedef struct _GstVideoDiffClass GstVideoDiffClass;

struct _GstVideoDiff
{
  GstVideoFilter base_videodiff;

  GstBuffer *previous_buffer;
  GstVideoInfo oldinfo;

  int threshold;
  int t;
};

struct _GstVideoDiffClass
{
  GstVideoFilterClass base_videodiff_class;
};

GType gst_video_diff_get_type (void);
GST_ELEMENT_REGISTER_DECLARE (videodiff);

G_END_DECLS

#endif
